@@ -16,18 +16,18 @@
  *
  **********************************************************************/
 
-#ifndef GEOS_GEOM_ENVELOPE_H
-#define GEOS_GEOM_ENVELOPE_H
+#pragma once
 
 
 #include <geos/export.h>
-#include <geos/inline.h>
 #include <geos/geom/Coordinate.h>
 
 #include <string>
 #include <vector>
 #include <ostream> // for operator<<
 #include <memory>
+#include <cassert>
+#include <algorithm>
 
 namespace geos {
 namespace geom { // geos::geom
@@ -66,7 +66,12 @@ public:
     /** \brief
      * Creates a null Envelope.
      */
-    Envelope();
+    Envelope()
+        : minx(DoubleNotANumber)
+        , maxx(DoubleNotANumber)
+        , miny(DoubleNotANumber)
+        , maxy(DoubleNotANumber)
+        {};
 
     /** \brief
      * Creates an Envelope for a region defined by maximum and minimum values.
@@ -76,7 +81,10 @@ public:
      * @param  y1  the first y-value
      * @param  y2  the second y-value
      */
-    Envelope(double x1, double x2, double y1, double y2);
+    Envelope(double x1, double x2, double y1, double y2)
+    {
+        init(x1, x2, y1, y2);
+    }
 
     /** \brief
      * Creates an Envelope for a region defined by two Coordinates.
@@ -84,26 +92,29 @@ public:
      * @param  p1  the first Coordinate
      * @param  p2  the second Coordinate
      */
-    Envelope(const Coordinate& p1, const Coordinate& p2);
+    Envelope(const CoordinateXY& p1, const CoordinateXY& p2)
+    {
+        init(p1, p2);
+    }
 
     /** \brief
      * Creates an Envelope for a region defined by a single Coordinate.
      *
      * @param  p  the Coordinate
      */
-    Envelope(const Coordinate& p);
-
-    /// Copy constructor
-    Envelope(const Envelope& env);
-
-    /// Assignment operator
-    Envelope& operator=(const Envelope& e);
+    explicit Envelope(const CoordinateXY& p)
+        : minx(p.x)
+        , maxx(p.x)
+        , miny(p.y)
+        , maxy(p.y)
+    {
+    }
 
     /** \brief
      * Create an Envelope from an Envelope string representation produced
      * by Envelope::toString()
      */
-    Envelope(const std::string& str);
+    explicit Envelope(const std::string& str);
 
     /** \brief
      * Test the point `q` to see whether it intersects the Envelope
@@ -114,8 +125,8 @@ public:
      * @param q the point to test for intersection
      * @return `true` if q intersects the envelope p1-p2
      */
-    static bool intersects(const Coordinate& p1, const Coordinate& p2,
-                           const Coordinate& q);
+    static bool intersects(const CoordinateXY& p1, const CoordinateXY& p2,
+                           const CoordinateXY& q);
 
     /** \brief
      * Test the envelope defined by `p1-p2` for intersection
@@ -128,8 +139,32 @@ public:
      *
      * @return `true` if Q intersects P
      */
-    static bool intersects(const Coordinate& p1, const Coordinate& p2,
-                           const Coordinate& q1, const Coordinate& q2);
+    static bool intersects(
+        const CoordinateXY& p1, const CoordinateXY& p2,
+        const CoordinateXY& q1, const CoordinateXY& q2)
+    {
+        double minq = std::min(q1.x, q2.x);
+        double maxq = std::max(q1.x, q2.x);
+        double minp = std::min(p1.x, p2.x);
+        double maxp = std::max(p1.x, p2.x);
+        if(minp > maxq) {
+            return false;
+        }
+        if(maxp < minq) {
+            return false;
+        }
+        minq = std::min(q1.y, q2.y);
+        maxq = std::max(q1.y, q2.y);
+        minp = std::min(p1.y, p2.y);
+        maxp = std::max(p1.y, p2.y);
+        if(minp > maxq) {
+            return false;
+        }
+        if(maxp < minq) {
+            return false;
+        }
+        return true;
+    }
 
     /** \brief
      * Check if the extent defined by two extremal points intersects
@@ -139,12 +174,15 @@ public:
      * @param b another point
      * @return `true` if the extents intersect
      */
-    bool intersects(const Coordinate& a, const Coordinate& b) const;
+    bool intersects(const CoordinateXY& a, const CoordinateXY& b) const;
 
     /** \brief
      *  Initialize to a null Envelope.
      */
-    void init(void);
+    void init()
+    {
+        setToNull();
+    };
 
     /** \brief
      * Initialize an Envelope for a region defined by maximum and minimum values.
@@ -154,7 +192,25 @@ public:
      * @param  y1  the first y-value
      * @param  y2  the second y-value
      */
-    void init(double x1, double x2, double y1, double y2);
+    void init(double x1, double x2, double y1, double y2)
+    {
+        if(x1 < x2) {
+            minx = x1;
+            maxx = x2;
+        }
+        else {
+            minx = x2;
+            maxx = x1;
+        }
+        if(y1 < y2) {
+            miny = y1;
+            maxy = y2;
+        }
+        else {
+            miny = y2;
+            maxy = y1;
+        }
+    };
 
     /** \brief
      * Initialize an Envelope to a region defined by two Coordinates.
@@ -162,23 +218,29 @@ public:
      * @param  p1  the first Coordinate
      * @param  p2  the second Coordinate
      */
-    void init(const Coordinate& p1, const Coordinate& p2);
+    void init(const CoordinateXY& p1, const CoordinateXY& p2)
+    {
+        init(p1.x, p2.x, p1.y, p2.y);
+    };
 
     /** \brief
      * Initialize an Envelope to a region defined by a single Coordinate.
      *
      * @param  p  the Coordinate
      */
-    void init(const Coordinate& p);
-
-    // use assignment operator instead
-    //void init(Envelope env);
+    void init(const CoordinateXY& p)
+    {
+        init(p.x, p.x, p.y, p.y);
+    };
 
     /** \brief
      * Makes this `Envelope` a "null" envelope, that is, the envelope
      * of the empty geometry.
      */
-    void setToNull(void);
+    void setToNull()
+    {
+        minx = maxx = miny = maxy = DoubleNotANumber;
+    };
 
     /** \brief
      * Returns `true` if this Envelope is a "null" envelope.
@@ -186,21 +248,36 @@ public:
      * @return `true` if this Envelope is uninitialized or is the
      *                envelope of the empty geometry.
      */
-    bool isNull(void) const;
+    bool isNull(void) const
+    {
+        return std::isnan(maxx);
+    };
 
     /** \brief
      * Returns the difference between the maximum and minimum x values.
      *
      * @return  `max x - min x`, or 0 if this is a null Envelope
      */
-    double getWidth(void) const;
+    double getWidth() const
+    {
+        if(isNull()) {
+            return 0;
+        }
+        return maxx - minx;
+    }
 
     /** \brief
      * Returns the difference between the maximum and minimum y values.
      *
      * @return `max y - min y`, or 0 if this is a null Envelope
      */
-    double getHeight(void) const;
+    double getHeight() const
+    {
+        if(isNull()) {
+            return 0;
+        }
+        return maxy - miny;
+    }
 
     /** \brief
      * Gets the area of this envelope.
@@ -214,29 +291,69 @@ public:
         return getWidth() * getHeight();
     }
 
+    /** \brief
+     * Returns true if this Envelope covers a finite region
+     */
+    bool
+    isFinite() const
+    {
+        return std::isfinite(getArea());
+    }
+
     /** \brief
      * Returns the Envelope maximum y-value. `min y > max y` indicates
      * that this is a null Envelope.
      */
-    double getMaxY() const;
+    double getMaxY() const
+    {
+        assert(!isNull());
+        return maxy;
+    };
 
     /** \brief
      * Returns the Envelope maximum x-value. `min x > max x` indicates
      * that this is a null Envelope.
      */
-    double getMaxX() const;
+    double getMaxX() const
+    {
+        assert(!isNull());
+        return maxx;
+    };
 
     /** \brief
      * Returns the Envelope minimum y-value. `min y > max y` indicates
      * that this is a null Envelope.
      */
-    double getMinY() const;
+    double getMinY() const
+    {
+        assert(!isNull());
+        return miny;
+    };
 
     /** \brief
      * Returns the Envelope minimum x-value. `min x > max x` indicates
      * that this is a null Envelope.
      */
-    double getMinX() const;
+    double getMinX() const
+    {
+        assert(!isNull());
+        return minx;
+    };
+
+    /**
+     * Gets the length of the diameter (diagonal) of the envelope.
+     *
+     * @return the diameter length
+     */
+    double getDiameter() const
+    {
+        if (isNull()) {
+            return 0.0;
+        }
+        double w = getWidth();
+        double h = getHeight();
+        return std::sqrt(w*w + h*h);
+    }
 
     /** \brief
      * Computes the coordinate of the centre of this envelope
@@ -245,7 +362,7 @@ public:
      * @param centre The coordinate to write results into
      * @return `false` if the center could not be found (null envelope).
      */
-    bool centre(Coordinate& centre) const;
+    bool centre(CoordinateXY& centre) const;
 
     /** \brief
      * Computes the intersection of two [Envelopes](@ref Envelope).
@@ -286,8 +403,7 @@ public:
     expandBy(double p_distance)
     {
         expandBy(p_distance, p_distance);
-    }
-
+    };
 
     /** \brief
      * Enlarges the boundary of the Envelope so that it contains p. Does
@@ -295,7 +411,10 @@ public:
      *
      * @param  p the Coordinate to include
      */
-    void expandToInclude(const Coordinate& p);
+    void expandToInclude(const CoordinateXY& p)
+    {
+        expandToInclude(p.x, p.y);
+    };
 
     /** \brief
      * Enlarges the boundary of the Envelope so that it contains (x,y).
@@ -307,7 +426,29 @@ public:
      * @param  y  the value to lower the minimum y
      *            to or to raise the maximum y to
      */
-    void expandToInclude(double x, double y);
+    void expandToInclude(double x, double y)
+    {
+        if(isNull()) {
+            minx = x;
+            maxx = x;
+            miny = y;
+            maxy = y;
+        }
+        else {
+            if(x < minx) {
+                minx = x;
+            }
+            if(x > maxx) {
+                maxx = x;
+            }
+            if(y < miny) {
+                miny = y;
+            }
+            if(y > maxy) {
+                maxy = y;
+            }
+        }
+    };
 
     /** \brief
      * Enlarges the boundary of the Envelope so that it contains `other`.
@@ -316,8 +457,34 @@ public:
      *
      * @param other the Envelope to merge with
      */
-    void expandToInclude(const Envelope* other);
-    void expandToInclude(const Envelope& other);
+    void expandToInclude(const Envelope* other)
+    {
+        if(isNull()) {
+            minx = other->minx;
+            maxx = other->maxx;
+            miny = other->miny;
+            maxy = other->maxy;
+        }
+        else {
+            if(std::isless(other->minx, minx)) {
+                minx = other->minx;
+            }
+            if(std::isgreater(other->maxx, maxx)) {
+                maxx = other->maxx;
+            }
+            if(std::isless(other->miny, miny)) {
+                miny = other->miny;
+            }
+            if(std::isgreater(other->maxy, maxy)) {
+                maxy = other->maxy;
+            }
+        }
+    };
+
+    void expandToInclude(const Envelope& other)
+    {
+        return expandToInclude(&other);
+    };
 
     /** \brief
      * Tests if the Envelope `other` lies wholly inside this Envelope
@@ -351,7 +518,7 @@ public:
      *         of this Envelope.
      */
     bool
-    contains(const Coordinate& p) const
+    contains(const CoordinateXY& p) const
     {
         return covers(p.x, p.y);
     }
@@ -375,10 +542,14 @@ public:
     /** \brief
      * Check if the point p intersects (lies inside) the region of this Envelope.
      *
-     * @param  p the Coordinate to be tested
+     * @param other the Coordinate to be tested
      * @return true if the point intersects this Envelope
      */
-    bool intersects(const Coordinate& p) const;
+    bool intersects(const CoordinateXY& other) const
+    {
+        return (std::islessequal(other.x, maxx) && std::isgreaterequal(other.x, minx) &&
+                std::islessequal(other.y, maxy) && std::isgreaterequal(other.y,  miny));
+    }
 
     /** \brief
      *  Check if the point (x, y) intersects (lies inside) the region of this Envelope.
@@ -387,7 +558,13 @@ public:
      * @param y the y-ordinate of the point
      * @return `true` if the point intersects this Envelope
      */
-    bool intersects(double x, double y) const;
+    bool intersects(double x, double y) const
+    {
+        return std::islessequal(x, maxx) &&
+               std::isgreaterequal(x, minx) &&
+               std::islessequal(y, maxy) &&
+               std::isgreaterequal(y, miny);
+    }
 
     /** \brief
      * Check if the region defined by other Envelope intersects the region of this Envelope.
@@ -395,20 +572,35 @@ public:
      * @param other the Envelope which this Envelope is being checked for intersection
      * @return true if the Envelopes intersects
      */
-    bool intersects(const Envelope* other) const;
+    bool intersects(const Envelope* other) const
+    {
+        return std::islessequal(other->minx, maxx) &&
+               std::isgreaterequal(other->maxx, minx) &&
+               std::islessequal(other->miny, maxy) &&
+               std::isgreaterequal(other->maxy, miny);
+    }
 
-    bool intersects(const Envelope& other) const;
+    bool intersects(const Envelope& other) const
+    {
+        return intersects(&other);
+    }
 
     /**
     * Tests if the region defined by other
     * is disjoint from the region of this Envelope
     *
     * @param other  the Envelope being checked for disjointness
-    * @return  true if the Envelopes are disjoint
+    * @return true if the Envelopes are disjoint
     */
-    bool disjoint(const Envelope* other) const;
+    bool disjoint(const Envelope& other) const
+    {
+        return !intersects(other);
+    }
 
-    bool disjoint(const Envelope& other) const;
+    bool disjoint(const Envelope* other) const
+    {
+        return !intersects(other);
+    }
 
     /** \brief
      * Tests if the given point lies in or on the envelope.
@@ -417,7 +609,12 @@ public:
      * @param y the y-coordinate of the point which this Envelope is being checked for containing
      * @return `true` if `(x, y)` lies in the interior or on the boundary of this Envelope.
      */
-    bool covers(double x, double y) const;
+    bool covers(double x, double y) const {
+        return std::isgreaterequal(x,  minx) &&
+               std::islessequal(x, maxx) &&
+               std::isgreaterequal(y, miny) &&
+               std::islessequal(y,  maxy);
+    }
 
     /** \brief
      * Tests if the given point lies in or on the envelope.
@@ -425,7 +622,10 @@ public:
      * @param p the point which this Envelope is being checked for containing
      * @return `true` if the point lies in the interior or on the boundary of this Envelope.
      */
-    bool covers(const Coordinate* p) const;
+    bool covers(const CoordinateXY* p) const
+    {
+        return covers(p->x, p->y);
+    }
 
     /** \brief
      * Tests if the Envelope `other` lies wholly inside this Envelope (inclusive of the boundary).
@@ -450,12 +650,19 @@ public:
      */
     bool equals(const Envelope* other) const;
 
+    /**
+     * Returns `true` if all the extents of the Envelope are finite and defined (not NaN)
+     *
+     * @return `true` if envelope has only finite/valid extents, `false` otherwise
+     */
+    bool isfinite() const;
+
     /** \brief
      * Returns a `string` of the form `Env[minx:maxx,miny:maxy]`.
      *
      * @return a `string` of the form `Env[minx:maxx,miny:maxy]`
      */
-    std::string toString(void) const;
+    std::string toString() const;
 
     /** \brief
      * Computes the distance between this and another Envelope.
@@ -463,7 +670,20 @@ public:
      * The distance between overlapping Envelopes is 0. Otherwise, the
      * distance is the Euclidean distance between the closest points.
      */
-    double distance(const Envelope& env) const;
+    double distance(const Envelope& env) const
+    {
+        return std::sqrt(distanceSquared(env));
+    }
+
+    /** \brief
+     * Computes the maximum distance between points in this and another Envelope.
+     */
+    double maxDistance(const Envelope& other) const
+    {
+        Coordinate p(std::min(minx, other.minx), std::min(miny, other.miny));
+        Coordinate q(std::max(maxx, other.maxx), std::max(maxy, other.maxy));
+        return p.distance(q);
+    }
 
     /** \brief
      * Computes the square of the distance between this and another Envelope.
@@ -471,7 +691,17 @@ public:
      * The distance between overlapping Envelopes is 0. Otherwise, the
      * distance is the Euclidean distance between the closest points.
      */
-    double distanceSquared(const Envelope& env) const;
+    double distanceSquared(const Envelope& env) const
+    {
+        double dx = std::max(0.0,
+            std::max(maxx, env.maxx) - std::min(minx, env.minx) - (maxx - minx) -
+            (env.maxx - env.minx));
+        double dy = std::max(0.0,
+            std::max(maxy, env.maxy) - std::min(miny, env.miny) - (maxy - miny) -
+            (env.maxy - env.miny));
+
+        return dx * dx + dy * dy;
+    };
 
     /** \brief
      * Computes the distance between one Coordinate and an Envelope
@@ -479,10 +709,16 @@ public:
      * used to define the envelope is not significant.
      *
      * @param c the coordinate to from which distance should be found
-     * @param p1 first coordinate defining an envelope
-     * @param p2 second coordinate defining an envelope.
+     * @param p0 first coordinate defining an envelope
+     * @param p1 second coordinate defining an envelope.
      */
-    static double distanceToCoordinate(const Coordinate & c, const Coordinate & p1, const Coordinate & p2);
+    static double distanceToCoordinate(
+        const CoordinateXY& c,
+        const CoordinateXY& p0,
+        const CoordinateXY& p1)
+    {
+        return std::sqrt(distanceSquaredToCoordinate(c, p0, p1));
+    };
 
     /** \brief
      * Computes the squared distance between one Coordinate and an Envelope
@@ -490,12 +726,66 @@ public:
      * used to define the envelope is not significant.
      *
      * @param c the coordinate to from which distance should be found
-     * @param p1 first coordinate defining an envelope
-     * @param p2 second coordinate defining an envelope.
+     * @param p0 first coordinate defining an envelope
+     * @param p1 second coordinate defining an envelope.
      */
-    static double distanceSquaredToCoordinate(const Coordinate & c, const Coordinate & p1, const Coordinate & p2);
+    static double distanceSquaredToCoordinate(
+        const CoordinateXY& c,
+        const CoordinateXY& p0,
+        const CoordinateXY& p1)
+    {
+        double xa = c.x - p0.x;
+        double xb = c.x - p1.x;
+        double ya = c.y - p0.y;
+        double yb = c.y - p1.y;
 
-    size_t hashCode() const;
+        // If sign of a and b are not the same, then Envelope spans c and distance is zero.
+        double dx = (std::signbit(xa) == std::signbit(xb)) * std::min(std::abs(xa), std::abs(xb));
+        double dy = (std::signbit(ya) == std::signbit(yb)) * std::min(std::abs(ya), std::abs(yb));
+
+        return dx*dx + dy*dy;
+    }
+
+    std::size_t hashCode() const
+    {
+        auto hash = std::hash<double>{};
+
+        //Algorithm from Effective Java by Joshua Bloch [Jon Aquino]
+        std::size_t result = 17;
+        result = 37 * result + hash(minx);
+        result = 37 * result + hash(maxx);
+        result = 37 * result + hash(miny);
+        result = 37 * result + hash(maxy);
+        return result;
+    }
+
+    struct GEOS_DLL HashCode
+    {
+        std::size_t operator()(const Envelope& e) const
+        {
+            return e.hashCode();
+        };
+    };
+
+    /// Checks if two Envelopes are equal (2D only check)
+    // GEOS_DLL bool operator==(const Envelope& a, const Envelope& b);
+    GEOS_DLL friend bool
+    operator==(const Envelope& a, const Envelope& b)
+    {
+        return a.equals(&b);
+    }
+
+    // GEOS_DLL bool operator!=(const Envelope& a, const Envelope& b);
+    GEOS_DLL friend bool
+    operator!=(const Envelope& a, const Envelope& b)
+    {
+        return !(a == b);
+    }
+
+    /// Strict weak ordering operator for Envelope
+    /// This is the C++ equivalent of JTS's compareTo
+    GEOS_DLL friend bool
+    operator< (const Envelope& a, const Envelope& b);
 
 private:
 
@@ -505,10 +795,15 @@ private:
      * This is a generic function that really belongs in a utility
      * file somewhere
      */
-    std::vector<std::string> split(const std::string& str,
+    static std::vector<std::string> split(const std::string& str,
                                    const std::string& delimiters = " ");
 
-    static double distance(double x0, double y0, double x1, double y1);
+    static double distance(double x0, double y0, double x1, double y1)
+    {
+        double dx = x1 - x0;
+        double dy = y1 - y0;
+        return std::sqrt(dx * dx + dy * dy);
+    }
 
     /// the minimum x-coordinate
     double minx;
@@ -523,14 +818,8 @@ private:
     double maxy;
 };
 
-/// Checks if two Envelopes are equal (2D only check)
-GEOS_DLL bool operator==(const Envelope& a, const Envelope& b);
+
+
 
 } // namespace geos::geom
 } // namespace geos
-
-#ifdef GEOS_INLINE
-# include "geos/geom/Envelope.inl"
-#endif
-
-#endif // ndef GEOS_GEOM_ENVELOPE_H
